# ferny - asyncio SSH client library, using ssh(1)
#
# Copyright (C) 2023 Allison Karlitskaya <allison.karlitskaya@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import array
import ast
import asyncio
import contextlib
import logging
import os
import re
import socket
import tempfile
from typing import Any, ClassVar, Dict, Generator, List, Optional, Sequence, TextIO, Tuple

from . import interaction_client

logger = logging.getLogger(__name__)


COMMAND_RE = re.compile(b'\0ferny\0([^\n]*)\0\0\n')
COMMAND_TEMPLATE = '\0ferny\0{(command, args)!r}\0\0\n'

BEIBOOT_GADGETS = {
    "command": fr"""
        import sys
        def command(command, *args):
            sys.stderr.write(f{COMMAND_TEMPLATE!r})
            sys.stderr.flush()
    """,
    "end": r"""
        def end():
            command('ferny.end')
    """,
}


class InteractionError(Exception):
    pass


try:
    recv_fds = socket.recv_fds
except AttributeError:
    # Python < 3.9

    def recv_fds(
        sock: socket.socket, bufsize: int, maxfds: int, flags: int = 0
    ) -> Tuple[bytes, List[int], int, None]:
        fds = array.array("i")
        msg, ancdata, flags, addr = sock.recvmsg(bufsize, socket.CMSG_LEN(maxfds * fds.itemsize))
        for cmsg_level, cmsg_type, cmsg_data in ancdata:
            if (cmsg_level == socket.SOL_SOCKET and cmsg_type == socket.SCM_RIGHTS):
                fds.frombytes(cmsg_data[:len(cmsg_data) - (len(cmsg_data) % fds.itemsize)])
        return msg, list(fds), flags, addr


def get_running_loop() -> asyncio.AbstractEventLoop:
    try:
        return asyncio.get_running_loop()
    except AttributeError:
        # Python 3.6
        return asyncio.get_event_loop()


# https://discuss.python.org/t/expanding-asyncio-support-for-socket-apis/19277/8
async def wait_readable(fd: int) -> None:
    loop = get_running_loop()
    future = loop.create_future()

    def _ready() -> None:
        if not future.cancelled():
            future.set_result(None)
    loop.add_reader(fd, _ready)

    try:
        await future
    finally:
        loop.remove_reader(fd)


class InteractionHandler:
    commands: ClassVar[Sequence[str]]

    async def run_command(self, command: str, args: Tuple, fds: List[int], stderr: str) -> None:
        raise NotImplementedError


class InteractionResponder(InteractionHandler):
    commands: ClassVar[Sequence[str]] = ('ferny.askpass',)

    async def do_askpass(self, messages: str, prompt: str, hint: str) -> Optional[str]:
        """Prompt the user for an authentication or confirmation interaction.

        'messages' is data that was sent to stderr before the interaction was requested.
        'prompt' is the interaction prompt.

        The expected response type depends on hint:

            - "confirm": ask for permission, returning "yes" if accepted
                - example: authorizing agent operation

            - "none": show a request without need for a response
                - example: please touch your authentication token

            - otherwise: return a password or other form of text token
                - examples: enter password, unlock private key

        In any case, the function should properly handle cancellation.  For the
        "none" case, this will be the normal way to dismiss the dialog.
        """
        return None

    async def do_hostkey(self, reason: str, host: str, algorithm: str, key: str, fingerprint: str) -> bool:
        """Prompt the user for a decision regarding acceptance of a host key.

        The "reason" will be either "HOSTNAME" or "ADDRESS" (if `CheckHostIP` is enabled).

        The host, algorithm, and key parameters are the values in the form that
        they would appear one a single line in the known hosts file.  The
        fingerprint is the key fingerprint in the format that ssh would
        normally present it to the user.

        In case the host key should be accepted, this function needs to return
        True.  Returning False means that ssh implements its default logic.  To
        interrupt the connection, raise an exception.
        """
        return False

    async def do_custom_command(self, command: str, args: Tuple, fds: List[int], stderr: str) -> None:
        """Handle a custom command.

        The command name, its arguments, the passed fds, and the stderr leading
        up to the command invocation are all provided.

        See doc/interaction-protocol.md
        """

    async def _askpass_task(self, args: List[str], env: Dict[str, str],
                            status: TextIO, stdout: TextIO, stderr: str) -> None:
        logger.debug('_askpass_task(%s, %d, %s, %s, %s)', args, len(env), status, stdout, stderr)

        if len(args) == 2:
            # normal askpass
            answer = await self.do_askpass(stderr, args[1], env.get('SSH_ASKPASS_PROMPT', ''))
            if answer is not None:
                print(answer, file=stdout)
                print(0, file=status)

        elif len(args) == 6:
            # KnownHostsCommand
            argv0, reason, host, algorithm, key, fingerprint = args
            if reason in ['ADDRESS', 'HOSTNAME']:
                if await self.do_hostkey(reason, host, algorithm, key, fingerprint):
                    print(host, algorithm, key, file=stdout)
            print(0, file=status)

        else:
            logger.error('Incorrect number of command-line arguments to ferny-askpass: %s', args)

    async def _askpass_command(self, args: Tuple, fds: List[int], stderr: str) -> None:
        logger.debug('_askpass_command(%s, %s, %s)', args, fds, stderr)
        try:
            argv, env = args
            assert isinstance(argv, list)
            assert all(isinstance(arg, str) for arg in argv)
            assert isinstance(env, dict)
            assert all(isinstance(key, str) and isinstance(val, str) for key, val in env.items())
            assert len(fds) == 2
        except (ValueError, TypeError, AssertionError) as exc:
            logger.error('Invalid arguments to askpass interaction: %s, %s: %s', args, fds, exc)
            return

        with open(fds.pop(0), 'w') as status, open(fds.pop(0), 'w') as stdout:
            loop = get_running_loop()
            future = loop.create_future()

            # We want to wait until either of these things happen:
            #   - our handler function finishes running
            #   - status fd closes from the other side (ie: askpass was killed)
            def _done(task: Optional[asyncio.Task] = None) -> None:
                if not future.done():
                    future.set_result(None)

            task = loop.create_task(self._askpass_task(argv, env, status, stdout, stderr))
            task.add_done_callback(_done)
            loop.add_reader(status, _done)

            # We need to handle cancellation of our task — do our cleanup as a
            # finally: block.
            try:
                await future
            finally:
                # If the status fd closed first then we need to cancel the
                # askpass task.  In any case, we collect its result to make
                # sure any exceptions get propagated.
                loop.remove_reader(status)
                with contextlib.suppress(asyncio.CancelledError):
                    if not task.done():
                        task.cancel()
                    await task

    async def run_command(self, command: str, args: Tuple, fds: List[int], stderr: str) -> None:
        logger.debug('run_command(%s, %s, %s, %s)', command, args, fds, stderr)
        if command == 'ferny.askpass':
            await self._askpass_command(args, fds, stderr)
        else:
            await self.do_custom_command(command, args, fds, stderr)


class InteractionAgent:
    handlers: Dict[str, InteractionHandler]
    ours: socket.socket
    theirs: socket.socket
    buffer: bytes
    connected: bool

    def __init__(self, handler: Optional[InteractionHandler] = None) -> None:
        self.buffer = b''
        self.ours, self.theirs = socket.socketpair()
        self.connected = False
        self.handlers = {}

        if handler is not None:
            self.add_handler(handler)

    def add_handler(self, handler: InteractionHandler) -> None:
        for command in handler.commands:
            self.handlers[command] = handler

    def fileno(self) -> int:
        return self.theirs.fileno()

    async def invoke_command(self, stderr: bytes, command_blob: bytes, fds: List[int]) -> None:
        logger.debug('invoke_command(%s, %s, %s)', stderr, command_blob, fds)
        try:
            command, args = ast.literal_eval(command_blob.decode('utf-8'))
            if not isinstance(command, str) or not isinstance(args, tuple):
                raise TypeError('Invalid argument types')
        except (UnicodeDecodeError, SyntaxError, ValueError, TypeError) as exc:
            logger.error('Received invalid ferny command: %s: %s', command_blob, exc)
            return

        if command == 'ferny.end':
            logger.debug('  ferny.end -> setting connected=True')
            self.connected = True
            return

        try:
            handler = self.handlers[command]
        except KeyError:
            logger.error('Received unhandled ferny command: %s', command)
            return

        await handler.run_command(command, args, fds, stderr.decode('utf-8'))

    async def communicate(self) -> None:
        self.theirs.close()

        # Various bits of code call .pop() on the list to claim a particular
        # fd, but the ones that remain are our responsibility to close: we do
        # that at the end of each loop iteration, in the finally: block.
        fds: List[int] = []

        with self.ours:
            while not self.connected:
                try:
                    # Wait for a message to come in, and read it
                    await wait_readable(self.ours.fileno())
                    data, fds, _flags, _addr = recv_fds(self.ours, 4096, 10)
                    if not data:
                        raise InteractionError(self.buffer.decode('utf-8').strip())

                    # Add to the buffer
                    self.buffer += data

                    # Read zero or more "remote" messages
                    chunks = COMMAND_RE.split(self.buffer)
                    while len(chunks) > 1:
                        await self.invoke_command(chunks[0], chunks[1], [])
                        chunks = chunks[2:]
                    self.buffer = chunks[0]

                    # Maybe read one "local" message
                    if fds:
                        assert self.buffer.endswith(b'\0'), self.buffer
                        stderr = self.buffer[:-1]
                        self.buffer = b''
                        with open(fds.pop(0), 'rb') as command_channel:
                            command = command_channel.read()
                        await self.invoke_command(stderr, command, fds)

                finally:
                    while fds:
                        os.close(fds.pop())

        logger.debug('agent.communicate() complete.')


def write_askpass_to_tmpdir(tmpdir: str) -> str:
    askpass_path = os.path.join(tmpdir, 'ferny-askpass')
    fd = os.open(askpass_path, os.O_CREAT | os.O_WRONLY | os.O_CLOEXEC | os.O_EXCL | os.O_NOFOLLOW, 0o777)
    try:
        os.write(fd, __loader__.get_data(interaction_client.__file__))  # type: ignore
    finally:
        os.close(fd)
    return askpass_path


@contextlib.contextmanager
def temporary_askpass(**kwargs: Any) -> Generator[str, None, None]:
    with tempfile.TemporaryDirectory(**kwargs) as directory:
        yield write_askpass_to_tmpdir(directory)



import argparse
import asyncio
import json
import os
import sys
import tempfile
from pathlib import Path
from typing import Dict, Iterable

import pytest

from cockpit._vendor.systemd_ctypes import bus
from cockpit.bridge import Bridge
from cockpit.channels import CHANNEL_TYPES

from .mocktransport import MOCK_HOSTNAME, MockTransport


class test_iface(bus.Object):
    sig = bus.Interface.Signal('s')
    prop = bus.Interface.Property('s', value='none')

    @bus.Interface.Method('s')
    def get_prop(self) -> str:
        return self.prop


@pytest.fixture
def bridge(event_loop) -> Bridge:
    async def get_bridge() -> Bridge:
        bridge = Bridge(argparse.Namespace(privileged=False, beipack=False))
        asyncio.set_event_loop(None)
        # We use this for assertions
        bridge.superuser_bridges = list(bridge.superuser_rule.bridges)  # type: ignore[attr-defined]
        return bridge

    return event_loop.run_until_complete(get_bridge())


def add_pseudo(bridge: Bridge) -> None:
    bridge.more_superuser_bridges = [*bridge.superuser_bridges, 'pseudo']  # type: ignore[attr-defined]

    # Add pseudo to the existing set of superuser rules
    assert bridge.packages is not None
    configs = bridge.packages.get_bridge_configs()
    configs.append({
        'label': 'pseudo',
        'spawn': [
            sys.executable, os.path.abspath(f'{__file__}/../pseudo.py'),
            sys.executable, '-m', 'cockpit.bridge', '--privileged'
        ],
        'environ': [
            f'PYTHONPATH={":".join(sys.path)}'
        ],
        'privileged': True
    })
    bridge.superuser_rule.set_configs(configs)


@pytest.fixture
def no_init_transport(event_loop: asyncio.AbstractEventLoop,
                      bridge: Bridge) -> Iterable[MockTransport]:
    async def get_transport() -> MockTransport:
        return MockTransport(bridge)
    transport = event_loop.run_until_complete(get_transport())
    try:
        yield transport
    finally:
        event_loop.run_until_complete(transport.stop())


@pytest.fixture
def transport(event_loop: asyncio.AbstractEventLoop,
              no_init_transport: MockTransport) -> MockTransport:
    event_loop.run_until_complete(no_init_transport.assert_msg('', command='init'))
    no_init_transport.send_init()
    return no_init_transport


@pytest.mark.asyncio
async def test_echo(transport):
    echo = await transport.check_open('echo')

    transport.send_data(echo, b'foo')
    await transport.assert_data(echo, b'foo')

    transport.send_ping(channel=echo)
    await transport.assert_msg('', command='pong', channel=echo)

    transport.send_done(echo)
    await transport.assert_msg('', command='done', channel=echo)
    await transport.assert_msg('', command='close', channel=echo)


@pytest.mark.asyncio
async def test_host(transport):
    # try to open a null channel, explicitly naming our host
    await transport.check_open('null', host=MOCK_HOSTNAME)

    # try to open a null channel, no host
    await transport.check_open('null')

    # try to open a null channel, a different host which is sure to fail DNS
    await transport.check_open('null', host='¡invalid!', problem='no-host')

    # make sure host check happens before superuser
    # ie: requesting superuser=True on another host should fail because we
    # can't contact the other host ('no-host'), rather than trying to first
    # go to our superuser bridge ('access-denied')
    await transport.check_open('null', host='¡invalid!', superuser=True, problem='no-host')

    # but make sure superuser is indeed failing as we expect, on our host
    await transport.check_open('null', host=MOCK_HOSTNAME, superuser=True, problem='access-denied')


@pytest.mark.asyncio
async def test_dbus_call_internal(bridge, transport):
    my_object = test_iface()
    bridge.internal_bus.export('/foo', my_object)
    assert my_object._dbus_bus == bridge.internal_bus.server
    assert my_object._dbus_path == '/foo'

    values, = await transport.check_bus_call('/foo', 'org.freedesktop.DBus.Properties',
                                             'GetAll', ["test.iface"])
    assert values == {'Prop': {'t': 's', 'v': 'none'}}

    result, = await transport.check_bus_call('/foo', 'test.iface', 'GetProp', [])
    assert result == 'none'


@pytest.mark.asyncio
async def test_dbus_watch(bridge, transport):
    my_object = test_iface()
    bridge.internal_bus.export('/foo', my_object)
    assert my_object._dbus_bus == bridge.internal_bus.server
    assert my_object._dbus_path == '/foo'

    # Add a watch
    internal = await transport.ensure_internal_bus()

    transport.send_json(internal, watch={'path': '/foo', 'interface': 'test.iface'}, id='4')
    meta = await transport.next_msg(internal)
    assert meta['meta']['test.iface'] == {
        'methods': {'GetProp': {'in': [], 'out': ['s']}},
        'properties': {'Prop': {'flags': 'r', 'type': 's'}},
        'signals': {'Sig': {'in': ['s']}}
    }
    notify = await transport.next_msg(internal)
    assert notify['notify']['/foo'] == {'test.iface': {'Prop': 'none'}}
    reply = await transport.next_msg(internal)
    assert reply == {'id': '4', 'reply': []}

    # Change a property
    my_object.prop = 'xyz'
    notify = await transport.next_msg(internal)
    assert 'notify' in notify
    assert notify['notify']['/foo'] == {'test.iface': {'Prop': 'xyz'}}


async def verify_root_bridge_not_running(bridge, transport):
    assert bridge.superuser_rule.peer is None
    await transport.assert_bus_props('/superuser', 'cockpit.Superuser',
                                     {'Bridges': bridge.more_superuser_bridges, 'Current': 'none'})
    null = await transport.check_open('null', superuser=True, problem='access-denied')
    assert null not in bridge.open_channels


@pytest.mark.asyncio
async def verify_root_bridge_running(bridge, transport):
    await transport.assert_bus_props('/superuser', 'cockpit.Superuser',
                                     {'Bridges': bridge.more_superuser_bridges, 'Current': 'pseudo'})
    assert bridge.superuser_rule.peer is not None

    # try to open dbus on the root bridge
    root_dbus = await transport.check_open('dbus-json3', bus='internal', superuser=True)

    # verify that the bridge thinks that it's the root bridge
    await transport.assert_bus_props('/superuser', 'cockpit.Superuser',
                                     {'Bridges': [], 'Current': 'root'}, bus=root_dbus)

    # close up
    await transport.check_close(channel=root_dbus)


@pytest.mark.asyncio
async def test_superuser_dbus(bridge, transport):
    add_pseudo(bridge)
    await verify_root_bridge_not_running(bridge, transport)

    # start the superuser bridge -- no password, so it should work straight away
    () = await transport.check_bus_call('/superuser', 'cockpit.Superuser', 'Start', ['pseudo'])

    await verify_root_bridge_running(bridge, transport)

    # open a channel on the root bridge
    root_null = await transport.check_open('null', superuser=True)

    # stop the bridge
    stop = transport.send_bus_call(transport.internal_bus, '/superuser',
                                   'cockpit.Superuser', 'Stop', [])

    # that should have implicitly closed the open channel
    await transport.assert_msg('', command='close', channel=root_null)
    assert root_null not in bridge.open_channels

    # The Stop method call is done now
    await transport.assert_msg(transport.internal_bus, reply=[[]], id=stop)


def format_methods(methods: Dict[str, str]):
    return {name: {'t': 'a{sv}', 'v': {'label': {'t': 's', 'v': label}}} for name, label in methods.items()}


@pytest.mark.asyncio
async def test_superuser_dbus_pw(bridge, transport, monkeypatch):
    monkeypatch.setenv('PSEUDO_PASSWORD', 'p4ssw0rd')
    add_pseudo(bridge)
    await verify_root_bridge_not_running(bridge, transport)

    # watch for signals
    await transport.add_bus_match('/superuser', 'cockpit.Superuser')
    await transport.watch_bus('/superuser', 'cockpit.Superuser',
                              {
                                  'Bridges': bridge.more_superuser_bridges,
                                  'Current': 'none',
                                  'Methods': format_methods({'pseudo': 'pseudo'}),
                              })

    # start the bridge.  with a password this is more complicated
    start = transport.send_bus_call(transport.internal_bus, '/superuser',
                                    'cockpit.Superuser', 'Start', ['pseudo'])
    # first, init state
    await transport.assert_bus_notify('/superuser', 'cockpit.Superuser', {'Current': 'init'})
    # then, we'll be asked for a password
    await transport.assert_bus_signal('/superuser', 'cockpit.Superuser', 'Prompt',
                                      ['', 'can haz pw?', '', False, ''])
    # give it
    await transport.check_bus_call('/superuser', 'cockpit.Superuser', 'Answer', ['p4ssw0rd'])
    # and now the bridge should be running
    await transport.assert_bus_notify('/superuser', 'cockpit.Superuser', {'Current': 'pseudo'})

    # Start call is now done
    await transport.assert_bus_reply(start, [])

    # double-check
    await verify_root_bridge_running(bridge, transport)


@pytest.mark.asyncio
async def test_superuser_dbus_wrong_pw(bridge, transport, monkeypatch):
    monkeypatch.setenv('PSEUDO_PASSWORD', 'p4ssw0rd')
    add_pseudo(bridge)
    await verify_root_bridge_not_running(bridge, transport)

    # watch for signals
    await transport.add_bus_match('/superuser', 'cockpit.Superuser')
    await transport.watch_bus('/superuser', 'cockpit.Superuser',
                              {
                                  'Bridges': bridge.more_superuser_bridges,
                                  'Current': 'none',
                                  'Methods': format_methods({'pseudo': 'pseudo'}),
                              })

    # start the bridge.  with a password this is more complicated
    start = transport.send_bus_call(transport.internal_bus, '/superuser',
                                    'cockpit.Superuser', 'Start', ['pseudo'])
    # first, init state
    await transport.assert_bus_notify('/superuser', 'cockpit.Superuser', {'Current': 'init'})
    # then, we'll be asked for a password
    await transport.assert_bus_signal('/superuser', 'cockpit.Superuser', 'Prompt',
                                      ['', 'can haz pw?', '', False, ''])
    # give it
    await transport.check_bus_call('/superuser', 'cockpit.Superuser',
                                   'Answer', ['p5ssw0rd'])  # wrong password
    # pseudo fails after the first wrong attempt
    await transport.assert_bus_notify('/superuser', 'cockpit.Superuser', {'Current': 'none'})

    # Start call is now done and returned failure
    await transport.assert_bus_error(start, 'cockpit.Superuser.Error', 'pseudo says: Bad password')

    # double-check
    await verify_root_bridge_not_running(bridge, transport)


@pytest.mark.asyncio
async def test_superuser_init(bridge, no_init_transport):
    add_pseudo(bridge)
    await no_init_transport.assert_msg('', command='init')
    no_init_transport.send_init(superuser={"id": "pseudo"})
    transport = no_init_transport

    # this should work right away without auth
    await transport.assert_msg('', command='superuser-init-done')

    await verify_root_bridge_running(bridge, transport)


@pytest.mark.asyncio
async def test_superuser_init_pw(bridge, no_init_transport, monkeypatch):
    monkeypatch.setenv('PSEUDO_PASSWORD', 'p4ssw0rd')
    add_pseudo(bridge)
    await no_init_transport.assert_msg('', command='init')
    no_init_transport.send_init(superuser={"id": "pseudo"})
    transport = no_init_transport

    msg = await transport.assert_msg('', command='authorize')
    transport.send_json('', command='authorize', cookie=msg['cookie'], response='p4ssw0rd')

    # that should have worked
    await transport.assert_msg('', command='superuser-init-done')

    await verify_root_bridge_running(bridge, transport)


@pytest.mark.asyncio
async def test_no_login_messages(transport):
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["{}"])
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Dismiss', [], [])
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["{}"])


@pytest.fixture
def login_messages_envvar(monkeypatch):
    if sys.version_info < (3, 8):
        pytest.skip("os.memfd_create new in 3.8")
    fd = os.memfd_create('login messages')
    os.write(fd, b"msg")
    # this is questionable (since it relies on ordering of fixtures), but it works
    monkeypatch.setenv('COCKPIT_LOGIN_MESSAGES_MEMFD', str(fd))
    return None


@pytest.mark.asyncio
async def test_login_messages(login_messages_envvar, transport):
    del login_messages_envvar

    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["msg"])
    # repeated read should get the messages again
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["msg"])
    # ...but not after they were dismissed
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Dismiss', [], [])
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["{}"])
    # idempotency
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Dismiss', [], [])
    await transport.check_bus_call('/LoginMessages', 'cockpit.LoginMessages', 'Get', [], ["{}"])


@pytest.mark.asyncio
async def test_freeze(bridge, transport):
    koelle = await transport.check_open('echo')
    malle = await transport.check_open('echo')

    # send a bunch of data to frozen koelle
    bridge.open_channels[koelle].freeze_endpoint()
    transport.send_data(koelle, b'x1')
    transport.send_data(koelle, b'x2')
    transport.send_data(koelle, b'x3')
    transport.send_done(koelle)

    # malle never freezes
    transport.send_data(malle, b'yy')
    transport.send_done(malle)

    # unfreeze koelle
    bridge.open_channels[koelle].thaw_endpoint()

    # malle should have sent its messages first
    await transport.assert_data(malle, b'yy')
    await transport.assert_msg('', command='done', channel=malle)
    await transport.assert_msg('', command='close', channel=malle)

    # the data from koelle should still be in the right order, though
    await transport.assert_data(koelle, b'x1')
    await transport.assert_data(koelle, b'x2')
    await transport.assert_data(koelle, b'x3')
    await transport.assert_msg('', command='done', channel=koelle)
    await transport.assert_msg('', command='close', channel=koelle)


@pytest.mark.asyncio
async def test_internal_metrics(transport):
    metrics = [
        {"name": "cpu.core.user", "derive": "rate"},
        {"name": "memory.used"},
    ]
    interval = 100
    source = 'internal'

    await transport.check_open('metrics1', source=source, interval=interval, metrics=metrics)
    _, data = await transport.next_frame()
    # first message is always the meta message
    meta = json.loads(data)
    assert isinstance(meta['timestamp'], float)
    assert meta['interval'] == interval
    assert meta['source'] == source
    assert isinstance(meta['metrics'], list)
    instances = len([m['instances'] for m in meta['metrics'] if m['name'] == 'cpu.core.user'][0])

    # actual data
    _, data = await transport.next_frame()
    data = json.loads(data)
    # cpu.core.user instances should be the same as meta sent instances
    assert instances == len(data[0][0])
    # all instances should be False, as this is a rate
    assert not all(d for d in data[0][0])
    # memory.used should be an integer
    assert isinstance(data[0][1], int)


@pytest.mark.asyncio
async def test_fsread1_errors(transport):
    await transport.check_open('fsread1', path='/etc/shadow', problem='access-denied')
    await transport.check_open('fsread1', path='/', problem='internal-error',
                               reply_keys={'message': "[Errno 21] Is a directory: '/'"})


@pytest.mark.asyncio
async def test_fslist1_no_watch(transport):
    tempdir = tempfile.TemporaryDirectory()
    dir_path = Path(tempdir.name)

    # empty
    ch = transport.send_open('fslist1', path=str(dir_path), watch=False)
    await transport.assert_msg('', command='done', channel=ch)
    await transport.check_close(channel=ch)

    # create a file and a directory in some_dir
    Path(dir_path, 'somefile').touch()
    Path(dir_path, 'somedir').mkdir()

    ch = transport.send_open('fslist1', path=str(dir_path), watch=False)
    # don't assume any ordering
    msg1 = await transport.next_msg(ch)
    msg2 = await transport.next_msg(ch)
    if msg1['type'] == 'file':
        msg1, msg2 = msg2, msg1
    assert msg1 == {'event': 'present', 'path': 'somedir', 'type': 'directory'}
    assert msg2 == {'event': 'present', 'path': 'somefile', 'type': 'file'}

    await transport.assert_msg('', command='done', channel=ch)
    await transport.check_close(channel=ch)


@pytest.mark.asyncio
async def test_fslist1_notexist(transport):
    await transport.check_open(
        'fslist1', path='/nonexisting', watch=False,
        problem='not-found',
        reply_keys={'message': "[Errno 2] No such file or directory: '/nonexisting'"})


@pytest.mark.asyncio
@pytest.mark.parametrize('channeltype', CHANNEL_TYPES)
async def test_channel(channeltype, tmp_path):
    bridge = Bridge(argparse.Namespace(privileged=False, beipack=False))
    transport = MockTransport(bridge)
    await transport.assert_msg('', command='init')
    transport.send_init()

    payload = channeltype.payload
    args = dict(channeltype.restrictions)

    async def serve_page(reader, writer):
        while True:
            line = await reader.readline()
            if line.strip():
                print('HTTP Request line', line)
            else:
                break

        print('Sending HTTP reply')
        writer.write(b'HTTP/1.1 200 OK\r\n\r\nA document\r\n')
        await writer.drain()
        writer.close()

    srv = str(tmp_path / 'sock')
    await asyncio.start_unix_server(serve_page, srv)

    if payload == 'fslist1':
        args = {'path': '/', 'watch': False}
    elif payload == 'fsread1':
        args = {'path': '/etc/passwd'}
    elif payload == 'fsreplace1':
        args = {'path': 'tmpfile'}
    elif payload == 'fswatch1':
        args = {'path': '/etc'}
    elif payload == 'http-stream1':
        args = {'internal': 'packages', 'method': 'GET', 'path': '/manifests.js',
                'headers': {'X-Forwarded-Proto': 'http', 'X-Forwarded-Host': 'localhost'}}
    elif payload == 'http-stream2':
        args = {'method': 'GET', 'path': '/bzzt', 'unix': srv}
    elif payload == 'stream':
        if 'spawn' in args:
            args = {'spawn': ['cat']}
        else:
            args = {'unix': srv}
    elif payload == 'metrics1':
        args['metrics'] = [{'name': 'memory.free'}]
    elif payload == 'dbus-json3':
        if not os.path.exists('/run/dbus/system_bus_socket'):
            pytest.skip('no dbus')
    else:
        args = {}

    print('sending open', payload, args)
    ch = transport.send_open(payload, **args)
    saw_data = False

    while True:
        channel, msg = await transport.next_frame()
        print(channel, msg)
        if channel == '':
            control = json.loads(msg)
            assert control['channel'] == ch
            command = control['command']
            if command == 'done':
                saw_data = True
            elif command == 'ready':
                # If we get ready, it's our turn to send data first.
                # Hopefully we didn't receive any before.
                assert isinstance(bridge.open_channels[ch], channeltype)
                assert not saw_data
                break
            elif command == 'close':
                # If we got an immediate close message then it should be
                # because the channel sent data and finished, without error.
                assert 'problem' not in control
                assert saw_data
                return
            else:
                pytest.fail('unexpected event', (payload, args, control))
        else:
            saw_data = True

    # If we're here, it's our turn to talk.  Say nothing.
    print('sending done')
    transport.send_done(ch)

    if payload in ['dbus-json3', 'fswatch1', 'null']:
        transport.send_close(ch)

    while True:
        channel, msg = await transport.next_frame()
        print(channel, msg)
        if channel == '':
            control = json.loads(msg)
            command = control['command']
            if command == 'done':
                continue
            elif command == 'close':
                assert 'problem' not in control
                return



# systemd_ctypes
#
# Copyright (C) 2022 Allison Karlitskaya <allison.karlitskaya@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
import os


class negative_errno(ctypes.c_int):
    def errcheck(self, func, _args):
        result = self.value
        if result < 0:
            raise OSError(-result, f'{func.__name__}: {os.strerror(-result)}')
        return result


class utf8(ctypes.c_char_p):
    def __init__(self, value=None):
        if value is not None:
            value = value.encode('utf-8')
        super().__init__(value)

    def errcheck(self, func, _args):
        return self.value

    @classmethod
    def from_param(cls, value):
        return cls(value)

    @property
    def value(self):
        value = super().value
        if value is not None:
            value = value.decode('utf-8')
        return value


class boolint(ctypes.c_int):
    @property
    def value(self):
        # https://github.com/python/cpython/issues/73456
        return bool(ctypes.c_int.value.__get__(self))


def instancemethod(func):
    func.argtypes = [ctypes.c_void_p, *func.argtypes]

    def wrapper(*args):
        return func(*args)
    return wrapper


class librarywrapper:
    def __getattr__(self, name):
        # This is mostly to silence pylint complaining about unknown attributes
        raise AttributeError(f'{self.namespace}_{name} is not registered')

    @classmethod
    def dlopen(cls, soname):
        cls._library = ctypes.cdll.LoadLibrary(soname)

    @classmethod
    def register_reference_types(cls, type_names):
        for name in type_names:
            class pointer_type(ctypes.c_void_p):
                __qualname__ = f'{cls.__qualname__}.{name}_p'
                __module__ = cls.__module__
                _library = cls._library
                namespace = f'{cls.namespace}_{name}'

                @classmethod
                def register_methods(cls, methods, private=False):
                    prefix = '_' if private else ''
                    for decorator, restype, name, argtypes in methods:
                        func = getattr(cls._library, f'{cls.namespace}_{name}')
                        func.restype, func.argtypes = restype, argtypes
                        if hasattr(restype, 'errcheck'):
                            func.errcheck = func.restype.errcheck
                        setattr(cls, f'{prefix}{name}', decorator(func))

            pointer_type.register_methods([(instancemethod, pointer_type, 'ref', [])], private=True)
            pointer_type.__name__ = f'{name}_p'
            setattr(cls, pointer_type.__name__, pointer_type)

            class reference_type(pointer_type):
                __qualname__ = f'{cls.__qualname__}.{name}'
                __module__ = cls.__module__

                @classmethod
                def ref(cls, obj):
                    return cls(obj._ref().value)

                def __del__(self):
                    if self.value is not None:
                        self._unref()

            reference_type.register_methods([(instancemethod, None, 'unref', [])], private=True)
            reference_type.__name__ = name
            setattr(cls, reference_type.__name__, reference_type)



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

from .. import data
from ..channel import Channel

logger = logging.getLogger(__name__)


class PackagesChannel(Channel):
    payload = 'http-stream1'
    restrictions = [("internal", "packages")]

    headers = None
    protocol = None
    host = None
    origin = None
    out_headers = None
    options = None
    post = None

    def push_header(self, key, value):
        if self.out_headers is None:
            self.out_headers = {}
        self.out_headers[key] = value

    def http_ok(self, content_type, extra_headers=None):
        headers = {'Content-Type': content_type}
        if self.out_headers is not None:
            headers.update(self.out_headers)
        if extra_headers is not None:
            headers.update(extra_headers)
        self.send_message(status=200, reason='OK', headers={k: v for k, v in headers.items() if v is not None})

    def http_error(self, status, message):
        template = data.read_cockpit_data_file('fail.html')
        self.send_message(status=status, reason='ERROR', headers={'Content-Type': 'text/html; charset=utf-8'})
        self.send_data(template.replace(b'@@message@@', message.encode('utf-8')))

    def do_done(self):
        assert not self.post
        assert self.options['method'] == 'GET'
        path = self.options['path']

        self.headers = self.options['headers']
        self.protocol = self.headers['X-Forwarded-Proto']
        self.host = self.headers['X-Forwarded-Host']
        self.origin = f'{self.protocol}://{self.host}'

        self.router.packages.serve_file(path, self)
        self.done()
        self.close()

    def do_data(self, data):
        self.post += data

    def do_open(self, options):
        self.post = b''
        self.options = options
        self.ready()



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import collections
import gzip
import hashlib
import json
import logging
import mimetypes
import os
import re
import shutil
from pathlib import Path
from typing import Callable, ClassVar, Dict, List, Optional, Pattern, Tuple

from cockpit._vendor.systemd_ctypes import bus

from . import config

VERSION = '300'
logger = logging.getLogger(__name__)


# An entity to serve is a set of bytes plus a content-type/content-encoding pair
Entity = Tuple[bytes, Tuple[Optional[str], Optional[str]]]
Entities = Dict[str, Entity]


# Sorting is important because the checksums are dependent on the order we
# visit these.  This is not the same as sorting on the full pathname, since
# each directory component is considered separately.  We could split the path
# and sort on that, but it ends up not being particularly elegant.
# Changing the approach here will change the checksum.
# We're allowed to change it as we wish, but let's try to match behaviour with
# cockpitpackages.c for the time being, since the unit tests depend on it.
def directory_items(path):
    return sorted(path.iterdir(), key=lambda item: item.name)


# HACK: We eventually want to get rid of all ${libexecdir} in manifests:
# - rewrite cockpit-{ssh,pcp} in Python and import them as module instead of exec'ing
# - rewrite cockpit-askpass in Python and write it into a temporary file
# - bundle helper shell scripts
# Until then, we need this libexecdir detection hack.
LIBEXECDIR = None


def get_libexecdir() -> str:
    """Detect libexecdir on current machine

    This only works for systems which have cockpit-ws installed.
    """
    global LIBEXECDIR
    if LIBEXECDIR is None:
        for candidate in ['/usr/local/libexec', '/usr/libexec', '/usr/local/lib/cockpit', '/usr/lib/cockpit']:
            if os.path.exists(os.path.join(candidate, 'cockpit-askpass')):
                LIBEXECDIR = candidate
                break
        else:
            logger.warning('Could not detect libexecdir')
            # give readable error messages
            LIBEXECDIR = '/nonexistent/libexec'

    return LIBEXECDIR


def patch_libexecdir(obj):
    if isinstance(obj, str):
        if '${libexecdir}/cockpit-askpass' in obj:
            # extra-special case: we handle this internally
            abs_askpass = shutil.which('cockpit-askpass')
            if abs_askpass is not None:
                return obj.replace('${libexecdir}/cockpit-askpass', abs_askpass)
        return obj.replace('${libexecdir}', get_libexecdir())
    elif isinstance(obj, dict):
        return {key: patch_libexecdir(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        return [patch_libexecdir(item) for item in obj]
    else:
        return obj


def parse_accept_language(headers: Dict[str, object]) -> List[str]:
    """Parse the Accept-Language header, if it exists.

    Returns an ordered list of languages.

    https://tools.ietf.org/html/rfc7231#section-5.3.5
    """

    accept_language = headers.get('Accept-Language')
    if not isinstance(accept_language, str):
        return []

    accept_languages = accept_language.split(',')
    locales = []
    for language in accept_languages:
        language = language.strip()
        locale, _, weightstr = language.partition(';q=')
        weight = float(weightstr or 1)

        # Skip possible empty locales
        if not locale:
            continue

        # Locales are case-insensitive and we store our list in lowercase
        locales.append((locale.lower(), weight))

    return [locale for locale, _ in sorted(locales, key=lambda k: k[1], reverse=True)]


def find_translation(translations: Entities, locales: List[str]) -> Entity:
    # First check the locales that the user sent
    for locale in locales:
        translation = translations.get(locale)
        if translation is not None:
            return translation

    # Next, check the language-only versions of variant-specified locales
    for locale in locales:
        language, _, region = locale.partition('-')
        if not region:
            continue
        translation = translations.get(language)
        if translation:
            return translation

    # If nothing else worked, we always have English
    return translations['en']


class PackagesListener:
    def packages_loaded(self):
        """Called when the packages have been reloaded"""


class Package:
    CONDITIONS: ClassVar[Dict[str, Callable[[str], bool]]] = {
        'path-exists': os.path.exists,
        'path-not-exists': lambda p: not os.path.exists(p),
    }

    # For po.js files, the interesting part is the locale name
    PO_JS_RE: ClassVar[Pattern] = re.compile(r'po\.([^.]+)\.js(\.gz)?')

    # A built in base set of "English" translations
    PO_EN_JS: ClassVar[Entity] = (b'', mimetypes.guess_type('po.js'))
    BASE_TRANSLATIONS: ClassVar[Entities] = {'en': PO_EN_JS, 'en-us': PO_EN_JS}

    files: Entities
    translations: Entities

    def __init__(self, path):
        self.path = path

        with (self.path / 'manifest.json').open(encoding='utf-8') as manifest_file:
            self.manifest = json.load(manifest_file)

        self.try_override(self.path / 'override.json')
        self.try_override(config.lookup_config(f'{path.name}.override.json'))
        self.try_override(config.DOT_CONFIG_COCKPIT / f'{path.name}.override.json')

        # HACK: drop this after getting rid of ${libexecdir}, see above
        self.manifest = patch_libexecdir(self.manifest)

        self.name = self.manifest.get('name', path.name)
        self.priority = self.manifest.get('priority', 1)
        self.bridges = self.manifest.get('bridges', [])

        self.files = {}
        self.translations = dict(Package.BASE_TRANSLATIONS)

        self.version = Package.sortify_version(VERSION)

    def try_override(self, path: Path) -> None:
        try:
            with path.open(encoding='utf-8') as override_file:
                override = json.load(override_file)
            self.manifest = self.merge_patch(self.manifest, override)
        except FileNotFoundError:
            # This is the expected usual case
            pass
        except json.JSONDecodeError as exc:
            # User input error: report a warning
            logger.warning('%s: %s', path, exc)

    # https://www.rfc-editor.org/rfc/rfc7386
    @staticmethod
    def merge_patch(target: object, patch: object) -> object:
        # Loosely based on example code from the RFC
        if not isinstance(patch, dict):
            return patch

        # Always take a copy ('result') — we never modify the input ('target')
        result = dict(target if isinstance(target, dict) else {})
        for name, value in patch.items():
            if value is not None:
                result[name] = Package.merge_patch(result.get(name), value)
            else:
                result.pop(name)
        return result

    @staticmethod
    def sortify_version(version: str) -> str:
        """Convert a version string to a form that can be compared"""
        # 0-pad each numeric component.  Only supports numeric versions like 1.2.3.
        return '.'.join(part.zfill(8) for part in version.split('.'))

    def add_file(self, item: Path, checksums: List['hashlib._Hash']) -> None:
        rel = str(item.relative_to(self.path))
        guessed_type = mimetypes.guess_type(rel)

        with item.open('rb') as file:
            data = file.read()

        # Keep the file in memory to serve it later: if this is a 'po.*.js'
        # file then add it to the list of translations, by locale name.
        # Otherwise, add it to the normal files list (after stripping '.gz').
        po_match = Package.PO_JS_RE.fullmatch(rel)
        if po_match:
            locale = po_match.group(1)
            # Accept-Language is case-insensitive and uses '-' to separate variants
            lower_locale = locale.lower().replace('_', '-')
            self.translations[lower_locale] = data, guessed_type
        else:
            basename = rel[:-3] if rel.endswith('.gz') else rel
            self.files[basename] = data, guessed_type

        # Perform checksum calculation
        sha = hashlib.sha256(data).hexdigest()
        for context in checksums:
            context.update(f'{rel}\0{sha}\0'.encode('ascii'))

    def walk(self, checksums, path):
        for item in directory_items(path):
            if item.is_dir():
                self.walk(checksums, item)
            elif item.is_file():
                self.add_file(item, checksums)

    def check(self, at_least_prio):
        if 'requires' in self.manifest:
            requires = self.manifest['requires']
            if any(package != 'cockpit' for package in requires):
                return False

            if 'cockpit' in requires and self.version < Package.sortify_version(requires['cockpit']):
                return False

        if at_least_prio is not None:
            if 'priority' not in self.manifest:
                return False

            if self.manifest['priority'] <= at_least_prio:
                return False

        for condition in self.manifest.get('conditions', []):
            try:
                (predicate, value), = condition.items()
            except (AttributeError, ValueError):
                # ignore manifests with broken syntax
                logger.warning('invalid condition in %s: %s', self.path, condition)
                return False

            try:
                test_fn = Package.CONDITIONS[predicate]
            except KeyError:
                # do *not* ignore manifests with unknown predicates, for forward compatibility
                logger.warning('ignoring unknown predicate in %s: %s', self.path, predicate)
                continue

            if not test_fn(value):
                logger.info('Hiding package %s as its %s condition is not met', self.path, condition)
                return False

        return True

    def get_content_security_policy(self, origin):
        assert origin.startswith('http')
        origin_ws = origin.replace('http', 'ws', 1)

        # fix me: unit tests depend on the specific order
        policy = collections.OrderedDict({
            "default-src": f"'self' {origin}",
            "connect-src": f"'self' {origin} {origin_ws}",
            "form-action": f"'self' {origin}",
            "base-uri": f"'self' {origin}",
            "object-src": "'none'",
            "font-src": f"'self' {origin} data:",
            "img-src": f"'self' {origin} data:",
        })

        manifest_policy = self.manifest.get('content-security-policy', '')
        for item in manifest_policy.split(';'):
            item = item.strip()
            if item:
                key, _, value = item.strip().partition(' ')
                policy[key] = value

        return ' '.join(f'{k} {v};' for k, v in policy.items()) + ' block-all-mixed-content'

    def serve_file(self, path, channel):
        if path == 'po.js':
            # We do locale-dependent lookup only for /po.js.  This always succeeds.
            locales = parse_accept_language(channel.headers)
            data, (content_type, encoding) = find_translation(self.translations, locales)
        else:
            # Otherwise, just look up the file based on its path.  May raise KeyError.
            data, (content_type, encoding) = self.files[path]

        headers = {
            "Access-Control-Allow-Origin": channel.origin,
            "Content-Encoding": encoding,
        }
        if content_type is not None and content_type.startswith('text/html'):
            headers['Content-Security-Policy'] = self.get_content_security_policy(channel.origin)

        channel.http_ok(content_type, headers)
        channel.send_data(data)


# TODO: This doesn't yet implement the slighty complicated checksum
# scheme of the C bridge (see cockpitpackages.c) to support caching
# and reloading.

class Packages(bus.Object, interface='cockpit.Packages'):
    manifests = bus.Interface.Property('s', value="{}")

    listener: Optional[PackagesListener]
    packages: Dict[str, Package]
    checksum: str = ''

    def __init__(self, listener: Optional[PackagesListener] = None):
        super().__init__()

        self.listener = listener
        self.packages = {}
        self.load_packages()

        # Reloading the Shell in the browser should reload the
        # packages.  This is implemented by having the Shell call
        # reload_hint whenever it starts.  The first call of this
        # method in each session is ignored so that packages are not
        # loaded twice right after logging in.
        #
        self.saw_first_reload_hint = False

    def show(self):
        for name in sorted(self.packages):
            package = self.packages[name]
            menuitems = ''
            print(f'{name:20} {menuitems:40} {package.path}')
        if self.checksum:
            print(f'checksum = {self.checksum}')

    def try_packages_dir(self, path, checksums):
        try:
            items = directory_items(path)
        except FileNotFoundError:
            return

        for item in items:
            if item.is_dir():
                try:
                    package = Package(item)
                except (FileNotFoundError, json.JSONDecodeError):
                    # ignore packages with broken/empty manifests
                    continue

                if package.name in self.packages:
                    at_least_prio = self.packages[package.name].priority
                else:
                    at_least_prio = None

                if package.check(at_least_prio):
                    self.packages[package.name] = package
                    package.walk(checksums, package.path)

    def load_packages(self):
        checksums = []

        xdg_data_home = os.environ.get('XDG_DATA_HOME') or os.path.expanduser('~/.local/share')
        self.try_packages_dir(Path(xdg_data_home) / 'cockpit', checksums)

        # we only checksum the system content if there's no user content
        if not self.packages:
            checksums.append(hashlib.sha256())

        xdg_data_dirs = os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share')
        for xdg_dir in xdg_data_dirs.split(':'):
            self.try_packages_dir(Path(xdg_dir) / 'cockpit', checksums)

        if checksums:
            self.checksum = checksums[0].hexdigest()
        else:
            self.checksum = None

        self.manifests = json.dumps({name: package.manifest for name, package in self.packages.items()})

    def serve_manifests_js(self, channel):
        channel.http_ok('text/javascript')

        # Send the translations required for the manifest files, from each package
        locales = parse_accept_language(channel.headers)
        for name, package in self.packages.items():
            if name in ['static', 'base1']:
                continue
            # find_translation will always find at least 'en'
            data, (content_type, encoding) = find_translation(package.translations, locales)
            if encoding == 'gzip':
                data = gzip.decompress(data)
            channel.send_data(data)

        channel.send_data(("""
            (function (root, data) {
                if (typeof define === 'function' && define.amd) {
                    define(data);
                }

                if (typeof cockpit === 'object') {
                    cockpit.manifests = data;
                } else {
                    root.manifests = data;
                }
            }(this, """ + self.manifests + """))""").encode('ascii'))

    def serve_manifests_json(self, channel):
        channel.http_ok('application/json')
        channel.send_data(self.manifests.encode('ascii'))

    def serve_package_file(self, path, channel):
        package, _, package_path = path[1:].partition('/')
        try:
            self.packages[package].serve_file(package_path, channel)
        except KeyError:
            channel.http_error(404, "Not Found")

    def serve_checksum(self, channel):
        channel.http_ok('text/plain')
        channel.send_data(self.checksum.encode('ascii'))

    def serve_file(self, path, channel):
        assert path[0] == '/'

        # HACK: If the response is language specific, don't cache the file. Caching "po.js" breaks
        # changing the language in Chromium, as that does not respect `Vary: Cookie` properly.
        # See https://github.com/cockpit-project/cockpit/issues/8160
        # We also can't cache /manifests.js because it changes if packages are installed/removed.
        locale_specific = path.endswith('po.js') or path in ['/manifests.js', '/manifests.json']
        if self.checksum is not None and not locale_specific:
            channel.push_header('X-Cockpit-Pkg-Checksum', self.checksum)
        else:
            channel.push_header('Cache-Control', 'no-cache, no-store')

        if path == '/manifests.js':
            self.serve_manifests_js(channel)
        elif path == '/manifests.json':
            self.serve_manifests_json(channel)
        elif path == '/checksum':
            self.serve_checksum(channel)
        else:
            self.serve_package_file(path, channel)

    def get_bridge_configs(self):
        bridges = []
        for package in sorted(self.packages.values(), key=lambda package: -package.priority):
            bridges.extend(package.bridges)
        return bridges

    @bus.Interface.Method()
    def reload(self):
        self.packages = {}
        self.load_packages()
        if self.listener is not None:
            self.listener.packages_loaded()

    @bus.Interface.Method()
    def reload_hint(self):
        if self.saw_first_reload_hint:
            self.reload()
        self.saw_first_reload_hint = True



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import logging
import os
import subprocess
from typing import Dict

from ..channel import ChannelError, ProtocolChannel
from ..transports import SubprocessProtocol, SubprocessTransport

logger = logging.getLogger(__name__)


class SocketStreamChannel(ProtocolChannel):
    payload = 'stream'

    async def create_transport(self, loop: asyncio.AbstractEventLoop,
                               options: Dict[str, object]) -> asyncio.Transport:
        if 'unix' in options and 'port' in options:
            raise ChannelError('protocol-error', message='cannot specify both "port" and "unix" options')

        try:
            # Unix
            if 'unix' in options:
                path = options['unix']
                # TODO: generic JSON validation
                if not isinstance(path, str):
                    raise ChannelError('protocol-error', message='unix option must be a string')
                label = f'Unix socket {path}'
                transport, _ = await loop.create_unix_connection(lambda: self, path)

            # TCP
            elif 'port' in options:
                try:
                    port: int = int(options['port'])  # type: ignore[call-overload]
                except ValueError as exc:
                    raise ChannelError('protocol-error', message='invalid "port" option for stream channel') from exc
                host = options.get('address', 'localhost')
                # TODO: generic JSON validation
                if not isinstance(host, str):
                    raise ChannelError('protocol-error',
                                       message='"address" option for stream channel must be a string')
                label = f'TCP socket {host}:{port}'

                transport, _ = await loop.create_connection(lambda: self, host, port)
            else:
                raise ChannelError('protocol-error',
                                   message='no "port" or "unix" or other address option for channel')

            logger.debug('SocketStreamChannel: connected to %s', label)
        except OSError as error:
            logger.info('SocketStreamChannel: connecting to %s failed: %s', label, error)
            if isinstance(error, ConnectionRefusedError):
                problem = 'not-found'
            else:
                problem = 'terminated'
            raise ChannelError(problem, message=str(error)) from error
        self.close_on_eof()
        assert isinstance(transport, asyncio.Transport)
        return transport


class SubprocessStreamChannel(ProtocolChannel, SubprocessProtocol):
    payload = 'stream'
    restrictions = (('spawn', None),)

    def process_exited(self) -> None:
        self.close_on_eof()

    def _get_close_args(self) -> Dict[str, object]:
        assert isinstance(self._transport, SubprocessTransport)
        args: Dict[str, object] = {'exit-status': self._transport.get_returncode()}
        stderr = self._transport.get_stderr()
        if stderr is not None:
            args['message'] = stderr
        return args

    def do_options(self, options):
        window = options.get('window')
        if window is not None:
            self._transport.set_window_size(**window)

    async def create_transport(self, loop: asyncio.AbstractEventLoop,
                               options: Dict[str, object]) -> SubprocessTransport:
        args = options['spawn']

        # TODO: generic JSON validation
        if not isinstance(args, list) or not all(isinstance(a, str) for a in args):
            raise ChannelError('protocol-error', message='invalid "args" option for stream channel')

        err = options.get('err')
        if err is not None and not isinstance(err, str):
            raise ChannelError('protocol-error', message='invalid "err" option for stream channel')

        cwd = options.get('directory')
        if cwd is not None and not isinstance(cwd, str):
            raise ChannelError('protocol-error', message='invalid "cwd" option for stream channel')

        pty = options.get('pty', False)
        if not isinstance(pty, bool):
            raise ChannelError('protocol-error', message='invalid "pty" option for stream channel')

        window = options.get('window')
        if window is not None and (
                not isinstance(window, dict) or
                not all(isinstance(k, str) and isinstance(v, int) for k, v in window.items())):
            raise ChannelError('protocol-error', message='invalid "window" option for stream channel')

        environ = options.get('environ')

        if err == 'out':
            stderr = subprocess.STDOUT
        elif err == 'ignore':
            stderr = subprocess.DEVNULL
        else:
            stderr = subprocess.PIPE

        env: Dict[str, str] = dict(os.environ)
        if environ is not None:
            if not isinstance(environ, list) or not all(isinstance(e, str) and '=' in e for e in environ):
                raise ChannelError('protocol-error', message='invalid "environ" option for stream channel')

            env.update(dict(e.split('=', 1) for e in environ))

        try:
            transport = SubprocessTransport(loop, self, args, pty=pty, window=window, env=env, cwd=cwd, stderr=stderr)
            logger.debug('Spawned process args=%s pid=%i', args, transport.get_pid())
            return transport
        except FileNotFoundError as error:
            raise ChannelError('not-found') from error
        except PermissionError as error:
            raise ChannelError('access-denied') from error
        except OSError as error:
            logger.info("Failed to spawn %s: %s", args, str(error))
            raise ChannelError('internal-error') from error



import glob
import os
import subprocess
import sys
from typing import Iterable

import pytest

SRCDIR = os.path.realpath(f'{__file__}/../../..')
BUILDDIR = os.environ.get('abs_builddir', SRCDIR)

SKIP = {
    'base1/test-dbus-address.html',
}

XFAIL = {
    'base1/test-websocket.html',
}


# Changed in version 3.10: Added the root_dir and dir_fd parameters.
def glob_py310(fnmatch: str, root_dir: str) -> Iterable[str]:
    prefix = f'{root_dir}/'
    prefixlen = len(prefix)

    for result in glob.glob(f'{prefix}{fnmatch}'):
        assert result.startswith(prefix)
        yield result[prefixlen:]


@pytest.mark.parametrize('html', glob_py310('*/test-*.html', root_dir=f'{SRCDIR}/qunit'))
def test_browser(html):
    if not os.path.exists(f'{BUILDDIR}/test-server'):
        pytest.skip('no test-server')
    if html in SKIP:
        pytest.skip()
    elif html in XFAIL:
        pytest.xfail()

    if 'COVERAGE_RCFILE' in os.environ:
        coverage = ['coverage', 'run', '--parallel-mode', '--module']
    else:
        coverage = []

    # Merge 2>&1 so that pytest displays an interleaved log
    subprocess.run(['test/common/tap-cdp', f'{BUILDDIR}/test-server',
                    sys.executable, '-m', *coverage, 'cockpit.bridge', '--debug',
                    f'./qunit/{html}'], check=True, stderr=subprocess.STDOUT)



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse
import asyncio
import json
import logging
import os
import pwd
import shlex
import socket
import subprocess
from typing import Dict, Iterable, List, Optional, Tuple, Type

from cockpit._vendor.ferny import interaction_client
from cockpit._vendor.systemd_ctypes import bus, run_async

from . import polyfills
from .channel import ChannelRoutingRule
from .channels import CHANNEL_TYPES
from .config import Config, Environment
from .internal_endpoints import EXPORTS
from .packages import Packages, PackagesListener
from .peer import PeersRoutingRule
from .remote import HostRoutingRule
from .router import Router
from .superuser import SuperuserRoutingRule
from .transports import StdioTransport

logger = logging.getLogger(__name__)


class InternalBus:
    exportees: List[bus.Slot]

    def __init__(self, exports: Iterable[Tuple[str, Type[bus.BaseObject]]]):
        client_socket, server_socket = socket.socketpair()
        self.client = bus.Bus.new(fd=client_socket.detach())
        self.server = bus.Bus.new(fd=server_socket.detach(), server=True)
        self.exportees = [self.server.add_object(path, cls()) for path, cls in exports]

    def export(self, path: str, obj: bus.BaseObject) -> None:
        self.exportees.append(self.server.add_object(path, obj))


class Bridge(Router, PackagesListener):
    internal_bus: InternalBus
    packages: Optional[Packages]
    bridge_rules: List[Dict[str, object]]
    args: argparse.Namespace

    def __init__(self, args: argparse.Namespace):
        self.internal_bus = InternalBus(EXPORTS)
        self.bridge_rules = []
        self.args = args

        self.superuser_rule = SuperuserRoutingRule(self, privileged=args.privileged)
        self.internal_bus.export('/superuser', self.superuser_rule)

        self.internal_bus.export('/config', Config())
        self.internal_bus.export('/environment', Environment())

        self.peers_rule = PeersRoutingRule(self)

        if args.beipack:
            # Some special stuff for beipack
            self.superuser_rule.set_configs([
                {
                    "privileged": True,
                    "spawn": ["sudo", "-k", "-A", "python3", "-ic", "# cockpit-bridge", "--privileged"],
                    "environ": ["SUDO_ASKPASS=ferny-askpass"],
                }
            ])
            self.packages = None
        elif args.privileged:
            self.packages = None
        else:
            self.packages = Packages(self)
            self.internal_bus.export('/packages', self.packages)
            self.packages_loaded()

        super().__init__([
            HostRoutingRule(self),
            self.superuser_rule,
            ChannelRoutingRule(self, CHANNEL_TYPES),
            self.peers_rule,
        ])

    @staticmethod
    def get_os_release():
        try:
            file = open('/etc/os-release', encoding='utf-8')
        except FileNotFoundError:
            try:
                file = open('/usr/lib/os-release', encoding='utf-8')
            except FileNotFoundError:
                logger.warning("Neither /etc/os-release nor /usr/lib/os-release exists")
                return {}

        with file:
            lexer = shlex.shlex(file, posix=True, punctuation_chars=True)
            return dict(token.split('=', 1) for token in lexer)

    def do_init(self, message: Dict[str, object]) -> None:
        superuser = message.get('superuser')
        if isinstance(superuser, dict):
            self.superuser_rule.init(superuser)

    def do_send_init(self) -> None:
        if self.packages is not None:
            packages_info = {
                'checksum': self.packages.checksum,
                'packages': {p: None for p in self.packages.packages}
            }
        else:
            packages_info = {}

        self.write_control(command='init', version=1, **packages_info,
                           os_release=self.get_os_release(),
                           capabilities={'explicit-superuser': True})

    # PackagesListener interface
    def packages_loaded(self):
        bridge_configs = self.packages.get_bridge_configs()
        if self.bridge_rules != bridge_configs:
            self.superuser_rule.set_configs(bridge_configs)
            self.peers_rule.set_configs(bridge_configs)
            self.bridge_configs = bridge_configs

    def eof_received(self) -> bool:
        # HACK: Make sure there's no outstanding sudo prompts blocking our shutdown
        self.superuser_rule.cancel_prompt()
        return super().eof_received()


async def run(args) -> None:
    logger.debug("Hi. How are you today?")

    # Unit tests require this
    me = pwd.getpwuid(os.getuid())
    os.environ['HOME'] = me.pw_dir
    os.environ['SHELL'] = me.pw_shell
    os.environ['USER'] = me.pw_name

    logger.debug('Starting the router.')
    router = Bridge(args)
    StdioTransport(asyncio.get_running_loop(), router)

    logger.debug('Startup done.  Looping until connection closes.')

    try:
        await router.communicate()
    except (BrokenPipeError, ConnectionResetError):
        # not unexpected if the peer doesn't hang up cleanly
        pass


def try_to_receive_stderr():
    try:
        ours, theirs = socket.socketpair()
        with ours:
            with theirs:
                interaction_client.command(2, 'cockpit.send-stderr', fds=[theirs.fileno()])
            _msg, fds, _flags, _addr = socket.recv_fds(ours, 1, 1)
    except OSError:
        return

    try:
        stderr_fd, = fds
        # We're about to abruptly drop our end of the stderr socketpair that we
        # share with the ferny agent.  ferny would normally treat that as an
        # unexpected error. Instruct it to do a clean exit, instead.
        interaction_client.command(2, 'ferny.end')
        os.dup2(stderr_fd, 2)
    finally:
        for fd in fds:
            os.close(fd)


def setup_logging(*, debug: bool):
    """Setup our logger with optional filtering of modules if COCKPIT_DEBUG env is set"""

    modules = os.getenv('COCKPIT_DEBUG', '')
    logging.basicConfig(format='%(name)s-%(levelname)s: %(message)s')

    if debug or modules == 'all':
        logging.getLogger().setLevel(level=logging.DEBUG)
    elif modules:
        for module in modules.split(','):
            module = module.strip()
            if not module:
                continue

            logging.getLogger(module).setLevel(logging.DEBUG)


def start_ssh_agent() -> None:
    # Launch the agent so that it goes down with us on EOF; PDEATHSIG would be more robust,
    # but it gets cleared on setgid ssh-agent, which some distros still do
    try:
        proc = subprocess.Popen(['ssh-agent', 'sh', '-ec', 'echo SSH_AUTH_SOCK=$SSH_AUTH_SOCK; read a'],
                                stdin=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
        assert proc.stdout is not None

        # Wait for the agent to write at least one line and look for the
        # listener socket.  If we fail to find it, kill the agent — something
        # went wrong.
        for token in shlex.shlex(proc.stdout.readline(), punctuation_chars=True):
            if token.startswith('SSH_AUTH_SOCK='):
                os.environ['SSH_AUTH_SOCK'] = token.replace('SSH_AUTH_SOCK=', '', 1)
                break
        else:
            proc.terminate()
            proc.wait()

    except FileNotFoundError:
        logger.debug("Couldn't start ssh-agent (FileNotFoundError)")

    except OSError as exc:
        logger.warning("Could not start ssh-agent: %s", exc)


def main(*, beipack: bool = False) -> None:
    polyfills.install()

    parser = argparse.ArgumentParser(description='cockpit-bridge is run automatically inside of a Cockpit session.')
    parser.add_argument('--privileged', action='store_true', help='Privileged copy of the bridge')
    parser.add_argument('--packages', action='store_true', help='Show Cockpit package information')
    parser.add_argument('--bridges', action='store_true', help='Show Cockpit bridges information')
    parser.add_argument('--rules', action='store_true', help='Show Cockpit bridge rules')
    parser.add_argument('--debug', action='store_true', help='Enable debug output (very verbose)')
    parser.add_argument('--version', action='store_true', help='Show Cockpit version information')
    args = parser.parse_args()

    # This is determined by who calls us
    args.beipack = beipack

    # If we were run with --privileged then our stderr is currently being
    # consumed by the main bridge looking for startup-related error messages.
    # Let's switch back to the original stderr stream, which has a side-effect
    # of indicating that our startup is more or less complete.  Any errors
    # after this point will land in the journal.
    if args.privileged:
        try_to_receive_stderr()

    setup_logging(debug=args.debug)

    # Special modes
    if args.packages:
        Packages().show()
        return
    elif args.bridges:
        print(json.dumps(Packages().get_bridge_configs(), indent=2))
        return

    # The privileged bridge doesn't need ssh-agent, but the main one does
    if 'SSH_AUTH_SOCK' not in os.environ and not args.privileged:
        start_ssh_agent()

    # asyncio.run() shim for Python 3.6 support
    run_async(run(args), debug=args.debug)


if __name__ == '__main__':
    main()



# systemd_ctypes
#
# Copyright (C) 2022 Allison Karlitskaya <allison.karlitskaya@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import selectors
import sys

from ctypes import byref

from . import inotify
from .libsystemd import sd


class Event(sd.event):
    _default = None

    @staticmethod
    def default():
        if not Event._default:
            Event._default = Event()
            sd.event.default(Event._default)
        return Event._default

    def add_inotify(self, path, mask, handler):
        @sd.event_inotify_handler_t
        def wrapper(source, _event, userdata):
            event = _event.contents
            handler(inotify.Event(event.mask), event.cookie, event.name if event.len else None)
            return 0
        source = sd.event_source()
        source.wrapper = wrapper
        super().add_inotify(byref(source), path, mask, source.wrapper, None)
        return source

    def add_inotify_fd(self, fd, mask, handler):
        # HACK: sd_event_add_inotify_fd() got added in 250, which is too new.  Fake it.
        return self.add_inotify(f'/proc/self/fd/{fd}', mask, handler)


# This is all a bit more awkward than it should have to be: systemd's event
# loop chaining model is designed for glib's prepare/check/dispatch paradigm;
# failing to call prepare() can lead to deadlocks, for example.
#
# Hack a selector subclass which calls prepare() before sleeping and this for us.
class Selector(selectors.DefaultSelector):
    def __init__(self, event=None):
        super().__init__()
        self.sd_event = event or Event.default()
        self.key = self.register(self.sd_event.get_fd(), selectors.EVENT_READ)

    def select(self, timeout=None):
        while self.sd_event.prepare():
            self.sd_event.dispatch()
        ready = super().select(timeout)
        # workaround https://github.com/systemd/systemd/issues/23826
        # keep calling wait() until there's nothing left
        while self.sd_event.wait(0):
            self.sd_event.dispatch()
            while self.sd_event.prepare():
                self.sd_event.dispatch()
        # NB: this could return zero events with infinite timeout, but nobody seems to mind.
        return [(key, events) for (key, events) in ready if key != self.key]


class EventLoopPolicy(asyncio.DefaultEventLoopPolicy):
    def new_event_loop(self):
        return asyncio.SelectorEventLoop(Selector())


def run_async(main, debug=None):
    asyncio.set_event_loop_policy(EventLoopPolicy())

    polyfill = sys.version_info < (3, 7, 0) and not hasattr(asyncio, 'run')
    if polyfill:
        # Polyfills for Python 3.6:
        loop = asyncio.get_event_loop()

        assert not hasattr(asyncio, 'get_running_loop')
        asyncio.get_running_loop = lambda: loop

        assert not hasattr(asyncio, 'create_task')
        asyncio.create_task = loop.create_task

        assert not hasattr(asyncio, 'run')

        def run(main, debug=None):
            if debug is not None:
                loop.set_debug(debug)
            loop.run_until_complete(main)

        asyncio.run = run

        asyncio._systemd_ctypes_polyfills = True

    asyncio.run(main, debug=debug)

    if polyfill:
        del asyncio.create_task, asyncio.get_running_loop, asyncio.run



from .errors import AuthenticationError, ChangedHostKeyError, HostKeyError, SshError, UnknownHostKeyError
from .interaction_agent import (
    BEIBOOT_GADGETS,
    COMMAND_TEMPLATE,
    InteractionAgent,
    InteractionError,
    InteractionHandler,
    InteractionResponder,
    temporary_askpass,
    write_askpass_to_tmpdir,
)
from .session import Session

__all__ = [
    'BEIBOOT_GADGETS',
    'COMMAND_TEMPLATE',
    'InteractionAgent',
    'InteractionError',
    'InteractionHandler',
    'InteractionResponder',
    'Session',
    'AuthenticationError',
    'HostKeyError',
    'ChangedHostKeyError',
    'UnknownHostKeyError',
    'SshError',
    'temporary_askpass',
    'write_askpass_to_tmpdir'
]



from .testlib import Browser

# used in cockpit-machines, cockpit-podman, cockpit-certificates
Browser.get_checked

# kept as being potentially useful in the future
Browser.upload_file
Browser.wait_attr_not_contains



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import http.client
import logging
import socket
import ssl
import threading

from ..channel import Channel

logger = logging.getLogger(__name__)


class HttpChannel(Channel):
    payload = 'http-stream2'

    def create_connection(self):
        opt_address = self.options.get('address') or 'localhost'
        opt_port = self.options.get('port')
        opt_unix = self.options.get('unix')
        opt_tls = self.options.get('tls')
        logger.debug('connecting to %s:%s; tls: %s', opt_address, opt_port or opt_unix, opt_tls)

        if opt_tls is not None:
            if 'authority' in opt_tls:
                if 'data' in opt_tls['authority']:
                    context = ssl.create_default_context(cadata=opt_tls['authority']['data'])
                else:
                    context = ssl.create_default_context(cafile=opt_tls['authority']['file'])
            else:
                context = ssl.create_default_context()

            if 'validate' in opt_tls and not opt_tls['validate']:
                context.check_hostname = False
                context.verify_mode = ssl.VerifyMode.CERT_NONE

            connection = http.client.HTTPSConnection(opt_address, opt_port, context=context)
        else:
            connection = http.client.HTTPConnection(opt_address, opt_port)

        try:
            if opt_unix:
                # create the connection's socket so that it won't call .connect() internally (which only supports TCP)
                connection.sock = socket.socket(socket.AF_UNIX)
                connection.sock.connect(opt_unix)
            else:
                # explicitly call connect(), so that we can do proper error handling
                connection.connect()
        except (OSError, IOError) as e:
            logger.error('Failed to open %s:%s: %s %s', opt_address, opt_port or opt_unix, type(e), e)
            problem = 'unknown-hostkey' if isinstance(e, ssl.SSLCertVerificationError) else 'not-found'
            self.close(problem=problem, message=str(e))
            return None

        return connection

    def read_send_response(self, response):
        """Completely read the response and send it to the channel"""

        while True:
            # we want to stream data blocks as soon as they come in
            block = response.read1(4096)
            if not block:
                logger.debug('reading response done')
                # this returns immediately and does not read anything more, but updates the http.client's
                # internal state machine to "response done"
                block = response.read()
                assert block == b''
                break
            logger.debug('read block of size %i', len(block))
            self.loop.call_soon_threadsafe(self.send_data, block)

    def parse_headers(self, http_msg):
        headers = dict(http_msg)
        remove = ['Connection', 'Transfer-Encoding']
        if self.options.get('binary'):
            remove = ['Content-Length', 'Range']
        for h in remove:
            try:
                del headers[h]
            except KeyError:
                pass
        return headers

    def request(self):
        connection = self.create_connection()
        if not connection:
            # make_connection does the error reporting
            return

        connection.request(self.options.get('method'),
                           self.options.get('path'),
                           headers=self.options.get('headers') or {},
                           body=self.body)
        try:
            response = connection.getresponse()
            self.loop.call_soon_threadsafe(lambda: self.send_control(
                command='response', status=response.status, reason=response.reason,
                headers=self.parse_headers(response.headers)))
            self.read_send_response(response)
        except (http.client.HTTPException, OSError) as error:
            msg = str(error)
            logger.debug('HTTP reading response failed: %s', msg)
            self.loop.call_soon_threadsafe(lambda: self.close(problem='terminated', message=msg))
            return
        finally:
            connection.close()

        self.loop.call_soon_threadsafe(self.done)
        self.loop.call_soon_threadsafe(self.close)
        logger.debug('closed')

    def do_open(self, options):
        logger.debug('open %s', options)
        # TODO: generic JSON validation
        if not options.get('method'):
            self.close(problem='protocol-error', message='missing or empty "method" field in HTTP stream request')
            return
        if options.get('path') is None:
            self.close(problem='protocol-error', message='missing "path" field in HTTP stream request')
            return
        if options.get('tls') is not None and options.get('unix'):
            self.close(problem='protocol-error', message='TLS on Unix socket is not supported')
            return
        if options.get('connection') is not None:
            self.close(problem='protocol-error', message='connection sharing is not implemented on this bridge')
            return

        opt_port = options.get('port')
        opt_unix = options.get('unix')
        if opt_port is None and opt_unix is None:
            self.close(problem='protocol-error', message='no "port" or "unix" option for channel')
            return
        if opt_port is not None and opt_unix is not None:
            self.close(problem='protocol-error', message='cannot specify both "port" and "unix" options')
            return

        self.options = options
        self.body = b''

        self.ready()

    def do_data(self, data):
        self.body += data

    def do_done(self):
        self.loop = asyncio.get_running_loop()
        threading.Thread(target=self.request, daemon=True).start()



# systemd_ctypes
#
# Copyright (C) 2022 Allison Karlitskaya <allison.karlitskaya@redhat.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

from ctypes import CFUNCTYPE, Structure, POINTER, byref, c_char, c_int, c_uint8, c_uint32, c_uint64, c_void_p

from .inotify import inotify_event
from .librarywrapper import librarywrapper, utf8, negative_errno, instancemethod, boolint


class sd(librarywrapper):
    namespace = 'sd'

    class bus_error(Structure):
        # This is ABI, so we are safe to assume it doesn't change.
        # Unfortunately, we lack anything like sd_bus_error_new().
        _fields_ = [
            ("name", utf8),
            ("message", utf8),
            ("_need_free", c_int)
        ]

        def get(self):
            return self.name.value, self.message.value

        def set(self, name, message):
            result = sd._library.sd_bus_error_set(byref(self), name, message)
            if result < 0:
                raise OSError(-result, f'sd_bus_error_set: {os.strerror(-result)}')

        def __del__(self):
            if self._b_needsfree_:
                sd._library.sd_bus_error_free(byref(self))

    class id128(Structure):
        # HACK: Pass-by-value of array-containing-structs is broken on Python
        # 3.6. See https://bugs.python.org/issue22273
        _fields_ = [
            ("bytes", c_uint8 * 16)
        ] if sys.version_info >= (3, 7, 0) else [
            ("one", c_uint64), ("two", c_uint64)
        ]


sd.dlopen('libsystemd.so.0')

sd.register_reference_types([
    'bus',
    'bus_message',
    'bus_slot',
    'event',
    'event_source',
])

sd.bus_message_handler_t = CFUNCTYPE(c_int, sd.bus_message_p, c_void_p, POINTER(sd.bus_error))
sd.event_inotify_handler_t = CFUNCTYPE(c_int, sd.event_source_p, POINTER(inotify_event), c_void_p)

sd.event.register_methods([
    (instancemethod, negative_errno, 'add_inotify',
     [POINTER(sd.event_source), utf8, c_uint32, sd.event_inotify_handler_t, c_void_p]),
    (instancemethod, negative_errno, 'dispatch', []),
    (instancemethod, negative_errno, 'get_fd', []),
    (instancemethod, negative_errno, 'get_state', []),
    (instancemethod, negative_errno, 'loop', []),
    (instancemethod, negative_errno, 'prepare', []),
    (instancemethod, negative_errno, 'wait', [c_uint64]),
    (staticmethod, negative_errno, 'default', [POINTER(sd.event_p)]),
])

sd.bus_message.register_methods([
    (instancemethod, negative_errno, 'append_basic', [sd.bus_message_p, c_char, c_void_p]),
    (instancemethod, negative_errno, 'read_basic', [sd.bus_message_p, c_char, c_void_p]),
    (instancemethod, POINTER(sd.bus_error), 'get_error', []),
    (instancemethod, negative_errno, 'at_end', [boolint]),
    (instancemethod, negative_errno, 'close_container', []),
    (instancemethod, negative_errno, 'enter_container', [c_char, utf8]),
    (instancemethod, negative_errno, 'exit_container', []),
    (instancemethod, negative_errno, 'has_signature', [utf8]),
    (instancemethod, negative_errno, 'is_method_error', [utf8]),
    (instancemethod, negative_errno, 'new_method_errnof', [POINTER(sd.bus_message_p), c_int, utf8, utf8]),
    (instancemethod, negative_errno, 'new_method_errorf', [POINTER(sd.bus_message_p), utf8, utf8, utf8]),
    (instancemethod, negative_errno, 'new_method_return', [POINTER(sd.bus_message_p)]),
    (instancemethod, negative_errno, 'open_container', [c_char, utf8]),
    (instancemethod, negative_errno, 'peek_type', [POINTER(c_char), POINTER(utf8)]),
    (instancemethod, negative_errno, 'rewind', [boolint]),
    (instancemethod, negative_errno, 'seal', [c_uint64, c_uint64]),
    (instancemethod, sd.bus_p, 'get_bus', []),
    (instancemethod, utf8, 'get_destination', []),
    (instancemethod, utf8, 'get_interface', []),
    (instancemethod, utf8, 'get_member', []),
    (instancemethod, utf8, 'get_path', []),
    (instancemethod, utf8, 'get_sender', []),
    (instancemethod, utf8, 'get_signature', [boolint]),
])

sd.bus.register_methods([
    (instancemethod, negative_errno, 'add_match', [POINTER(sd.bus_slot), utf8, sd.bus_message_handler_t, c_void_p]),
    (instancemethod, negative_errno, 'add_match_async',
     [POINTER(sd.bus_slot), utf8, sd.bus_message_handler_t, sd.bus_message_handler_t, c_void_p]),
    (instancemethod, negative_errno, 'add_object', [POINTER(sd.bus_slot), utf8, sd.bus_message_handler_t, c_void_p]),
    (instancemethod, negative_errno, 'attach_event', [sd.event_p, c_int]),
    (instancemethod, negative_errno, 'call',
     [sd.bus_message_p, c_uint64, POINTER(sd.bus_error), POINTER(sd.bus_message_p)]),
    (instancemethod, negative_errno, 'call_async',
     [POINTER(sd.bus_slot), sd.bus_message_p, sd.bus_message_handler_t, c_void_p, c_uint64]),
    (instancemethod, negative_errno, 'flush', []),
    (instancemethod, negative_errno, 'get_fd', []),
    (instancemethod, negative_errno, 'message_new', [POINTER(sd.bus_message_p), c_uint8]),
    (instancemethod, negative_errno, 'message_new_method_call', [POINTER(sd.bus_message_p), utf8, utf8, utf8, utf8]),
    (instancemethod, negative_errno, 'message_new_signal', [POINTER(sd.bus_message_p), utf8, utf8, utf8]),
    (instancemethod, negative_errno, 'new', [POINTER(sd.bus_p)]),
    (instancemethod, negative_errno, 'release_name', [utf8]),
    (instancemethod, negative_errno, 'request_name', [utf8, c_uint64]),
    (instancemethod, negative_errno, 'set_address', [utf8]),
    (instancemethod, negative_errno, 'set_bus_client', [boolint]),
    (instancemethod, negative_errno, 'set_fd', [c_int, c_int]),
    (instancemethod, negative_errno, 'set_server', [boolint, sd.id128]),
    (instancemethod, negative_errno, 'start', []),
    (instancemethod, negative_errno, 'wait', [c_uint64]),
    (instancemethod, negative_errno, 'send', [sd.bus_message_p, POINTER(c_uint64)]),
    (staticmethod, negative_errno, 'default_system', [POINTER(sd.bus_p)]),
    (staticmethod, negative_errno, 'default_user', [POINTER(sd.bus_p)]),
    (staticmethod, negative_errno, 'new', [POINTER(sd.bus_p)]),
])



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os
import random
from typing import Dict

from cockpit._vendor.systemd_ctypes import PathWatch
from cockpit._vendor.systemd_ctypes.inotify import Event as InotifyEvent

from ..channel import Channel, ChannelError, GeneratorChannel

logger = logging.getLogger(__name__)


def tag_from_stat(buf):
    return f'1:{buf.st_ino}-{buf.st_mtime}'


def tag_from_path(path):
    try:
        return tag_from_stat(os.stat(path))
    except FileNotFoundError:
        return '-'
    except OSError:
        return None


def tag_from_fd(fd):
    try:
        return tag_from_stat(os.fstat(fd))
    except OSError:
        return None


class FsListChannel(Channel):
    payload = 'fslist1'

    def send_entry(self, event, entry):
        if entry.is_symlink():
            mode = 'link'
        elif entry.is_file():
            mode = 'file'
        elif entry.is_dir():
            mode = 'directory'
        else:
            mode = 'special'

        self.send_message(event=event, path=entry.name, type=mode)

    def do_open(self, options):
        path = options.get('path')
        watch = options.get('watch', True)

        if watch:
            raise ChannelError('not-supported', message='watching is not implemented, use fswatch1')

        try:
            scan_dir = os.scandir(path)
        except OSError as error:
            if isinstance(error, FileNotFoundError):
                problem = 'not-found'
            elif isinstance(error, PermissionError):
                problem = 'access-denied'
            else:
                problem = 'internal-error'
            raise ChannelError(problem, message=str(error)) from error
        for entry in scan_dir:
            self.send_entry("present", entry)

        if not watch:
            self.done()
            self.close()


class FsReadChannel(GeneratorChannel):
    payload = 'fsread1'

    def do_yield_data(self, options: Dict[str, object]) -> GeneratorChannel.DataGenerator:
        binary = options.get('binary', False)
        max_read_size = options.get('max_read_size')
        # TODO: generic JSON validation
        if max_read_size is not None and not isinstance(max_read_size, int):
            raise ChannelError('protocol-error', message='max_read_size must be an integer')
        if not isinstance(options['path'], str):
            raise ChannelError('protocol-error', message='path is not a string')

        logger.debug('Opening file "%s" for reading', options['path'])

        try:
            with open(options['path'], 'rb') as filep:
                buf = os.stat(filep.fileno())
                if max_read_size is not None and buf.st_size > max_read_size:
                    raise ChannelError('too-large')

                while True:
                    data = filep.read1(Channel.BLOCK_SIZE)
                    if data == b'':
                        break
                    logger.debug('  ...sending %d bytes', len(data))
                    if not binary:
                        data = data.replace(b'\0', b'').decode('utf-8', errors='ignore').encode('utf-8')
                    yield data

            return {'tag': tag_from_stat(buf)}

        except FileNotFoundError:
            return {'tag': '-'}
        except PermissionError as exc:
            raise ChannelError('access-denied') from exc
        except OSError as exc:
            raise ChannelError('internal-error', message=str(exc)) from exc


class FsReplaceChannel(Channel):
    payload = 'fsreplace1'

    _path = None
    _tag = None
    _tempfile = None
    _temppath = None

    def unlink_temppath(self):
        try:
            os.unlink(self._temppath)
        except OSError:
            pass  # might have been removed from outside

    def do_open(self, options):
        self._path = options.get('path')
        self._tag = options.get('tag')
        self.ready()

    def do_data(self, data):
        if self._tempfile is None:
            # keep this bounded, in case anything unexpected goes wrong
            for _ in range(10):
                suffix = ''.join(random.choices("abcdefghijklmnopqrstuvwxyz0123456789_", k=6))
                self._temppath = f'{self._path}.cockpit-tmp.{suffix}'
                try:
                    fd = os.open(self._temppath, os.O_CREAT | os.O_WRONLY | os.O_EXCL, 0o666)
                    break
                except FileExistsError:
                    continue
                except PermissionError as exc:
                    raise ChannelError('access-denied') from exc
                except OSError as exc:
                    raise ChannelError('internal-error', message=str(exc)) from exc
            else:
                raise ChannelError('internal-error',
                                   message=f"Could not find unique file name for replacing {self._path}")

            try:
                self._tempfile = os.fdopen(fd, 'wb')
            except OSError:
                # Should Not Happen™, but let's be safe and avoid fd leak
                os.close(fd)
                self.unlink_temppath()
                raise

        self._tempfile.write(data)

    def do_done(self):
        if self._tempfile is None:
            try:
                os.unlink(self._path)
            # crash on other errors, as they are unexpected
            except FileNotFoundError:
                pass
        else:
            self._tempfile.flush()

            if self._tag and self._tag != tag_from_path(self._path):
                raise ChannelError('change-conflict')

            try:
                os.rename(self._temppath, self._path)
            except OSError:
                # ensure to not leave the temp file behind
                self.unlink_temppath()
                raise
            self._tempfile.close()
            self._tempfile = None

        self.done()
        self.close(tag=tag_from_path(self._path))

    def do_close(self):
        if self._tempfile is not None:
            self._tempfile.close()
            self.unlink_temppath()
            self._tempfile = None


class FsWatchChannel(Channel):
    payload = 'fswatch1'
    _tag = None
    _path = None
    _watch = None

    # The C bridge doesn't send the initial event, and the JS calls read()
    # instead to figure out the initial state of the file.  If we send the
    # initial state then we cause the event to get delivered twice.
    # Ideally we'll sort that out at some point, but for now, suppress it.
    _active = False

    @staticmethod
    def mask_to_event_and_type(mask):
        if (InotifyEvent.CREATE or InotifyEvent.MOVED_TO) in mask:
            return 'created', 'directory' if InotifyEvent.ISDIR in mask else 'file'
        elif InotifyEvent.MOVED_FROM in mask or InotifyEvent.DELETE in mask or InotifyEvent.DELETE_SELF in mask:
            return 'deleted', None
        elif InotifyEvent.ATTRIB in mask:
            return 'attribute-changed', None
        elif InotifyEvent.CLOSE_WRITE in mask:
            return 'done-hint', None
        else:
            return 'changed', None

    def do_inotify_event(self, mask, _cookie, name):
        logger.debug("do_inotify_event(%s): mask %X name %s", self._path, mask, name)
        event, type_ = self.mask_to_event_and_type(mask)
        if name:
            # file inside watched directory changed
            path = os.path.join(self._path, name.decode())
            tag = tag_from_path(path)
            self.send_message(event=event, path=path, tag=tag, type=type_)
        else:
            # the watched path itself changed; filter out duplicate events
            tag = tag_from_path(self._path)
            if tag == self._tag:
                return
            self._tag = tag
            self.send_message(event=event, path=self._path, tag=self._tag, type=type_)

    def do_identity_changed(self, fd, err):
        logger.debug("do_identity_changed(%s): fd %s, err %s", self._path, str(fd), err)
        self._tag = tag_from_fd(fd) if fd else '-'
        if self._active:
            self.send_message(event='created' if fd else 'deleted', path=self._path, tag=self._tag)

    def do_open(self, options):
        self._path = options['path']
        self._tag = None

        self._active = False
        self._watch = PathWatch(self._path, self)
        self._active = True

        self.ready()

    def do_close(self):
        self._watch.close()
        self._watch = None
        self.close()



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .dbus import DBusChannel
from .filesystem import FsListChannel, FsReadChannel, FsReplaceChannel, FsWatchChannel
from .http import HttpChannel
from .metrics import InternalMetricsChannel
from .packages import PackagesChannel
from .stream import SocketStreamChannel, SubprocessStreamChannel
from .trivial import EchoChannel, NullChannel

CHANNEL_TYPES = [
    DBusChannel,
    EchoChannel,
    FsListChannel,
    FsReadChannel,
    FsReplaceChannel,
    FsWatchChannel,
    HttpChannel,
    InternalMetricsChannel,
    NullChannel,
    PackagesChannel,
    SubprocessStreamChannel,
    SocketStreamChannel,
]



import asyncio
import os
import sys

import pytest

from cockpit.peer import ConfiguredPeer, PeerRoutingRule
from cockpit.protocol import CockpitProtocolError
from cockpit.router import Router

from . import mockpeer
from .mocktransport import MockTransport

PEER_CONFIG = {
    "spawn": [sys.executable, mockpeer.__file__],
    "environ": ['PYTHONPATH=' + ':'.join(sys.path)],
    "match": {"payload": "test"},
}


class Bridge(Router):
    init_host = 'localhost'

    def __init__(self):
        rule = PeerRoutingRule(self, PEER_CONFIG)
        super().__init__([rule])

    def do_send_init(self):
        pass


@pytest.fixture
def bridge():
    return Bridge()


@pytest.fixture
def transport(bridge):
    return MockTransport(bridge)


@pytest.fixture
def rule(bridge):
    return bridge.routing_rules[0]


@pytest.mark.asyncio
async def test_shutdown(transport, rule):
    await transport.check_open('test')
    await transport.check_open('xest', problem='not-supported')

    # Force the Peer closed
    rule.peer.close()
    await transport.assert_msg('', command='close', channel='channel.1', problem='terminated')

    # But it should spawn again
    await transport.check_open('test')
    await transport.check_open('xest', problem='not-supported')
    rule.peer.close()


@pytest.mark.asyncio
@pytest.mark.parametrize('init_type', ['wrong-command', 'channel-control', 'data', 'break-protocol'])
async def test_init_failure(rule, init_type, monkeypatch, transport):
    monkeypatch.setenv('INIT_TYPE', init_type)
    await transport.check_open('test', problem='protocol-error')


@pytest.mark.asyncio
async def test_immediate_shutdown(rule):
    peer = rule.apply_rule({'payload': 'test'})
    assert peer is not None
    peer.close()


@pytest.mark.asyncio
async def test_shutdown_before_init(monkeypatch, transport, rule):
    monkeypatch.setenv('INIT_TYPE', 'silence')
    channel = transport.send_open('test')
    assert rule.peer is not None
    assert rule.peer.transport is None
    while rule.peer.transport is None:
        await asyncio.sleep(0)
    rule.peer.close()
    await transport.assert_msg('', command='close', channel=channel, problem='terminated')


@pytest.mark.asyncio
async def test_exit_without_init(monkeypatch, transport):
    monkeypatch.setenv('INIT_TYPE', 'exit')
    await transport.check_open('test', problem='terminated')


@pytest.mark.asyncio
async def test_exit_not_found(monkeypatch, transport):
    monkeypatch.setenv('INIT_TYPE', 'exit-not-found')
    await transport.check_open('test', problem='no-cockpit')


@pytest.mark.asyncio
async def test_killed(monkeypatch, transport, rule):
    channel = await transport.check_open('test')
    os.kill(rule.peer.transport._process.pid, 9)
    await transport.assert_msg('', command='close', channel=channel, problem='terminated')


@pytest.mark.asyncio
@pytest.mark.parametrize('init_type', ['wrong-command', 'channel-control', 'data', 'break-protocol'])
async def test_await_failure(init_type, monkeypatch, bridge):
    monkeypatch.setenv('INIT_TYPE', init_type)
    peer = ConfiguredPeer(bridge, PEER_CONFIG)
    with pytest.raises(CockpitProtocolError):
        await peer.start()
    peer.close()


@pytest.mark.asyncio
async def test_await_broken_connect(bridge):
    class BrokenConnect(ConfiguredPeer):
        async def do_connect_transport(self):
            _ = 42 / 0

    peer = BrokenConnect(bridge, PEER_CONFIG)
    with pytest.raises(ZeroDivisionError):
        await peer.start()
    peer.close()


@pytest.mark.asyncio
async def test_await_broken_after_connect(bridge):
    class BrokenConnect(ConfiguredPeer):
        async def do_connect_transport(self):
            await super().do_connect_transport()
            _ = 42 / 0

    peer = BrokenConnect(bridge, PEER_CONFIG)
    with pytest.raises(ZeroDivisionError):
        await peer.start()
    peer.close()


class CancellableConnect(ConfiguredPeer):
    was_cancelled = False

    async def do_connect_transport(self):
        await super().do_connect_transport()
        try:
            # We should get cancelled here when the mockpeer sends "init"
            await asyncio.sleep(10000)
        except asyncio.CancelledError:
            self.was_cancelled = True
            raise


@pytest.mark.asyncio
async def test_await_cancellable_connect_init(bridge):
    peer = CancellableConnect(bridge, PEER_CONFIG)
    await peer.start()
    peer.close()
    while len(asyncio.all_tasks()) > 1:
        await asyncio.sleep(0.1)
    assert peer.was_cancelled


@pytest.mark.asyncio
async def test_await_cancellable_connect_close(monkeypatch, event_loop, bridge):
    monkeypatch.setenv('INIT_TYPE', 'silence')  # make sure we never get "init"
    peer = CancellableConnect(bridge, PEER_CONFIG)
    event_loop.call_later(0.1, peer.close)  # call peer.close() after .start() is running
    with pytest.raises(asyncio.CancelledError):
        await peer.start()
    # we already called .close()
    while len(asyncio.all_tasks()) > 1:
        await asyncio.sleep(0.1)
    assert peer.was_cancelled



# This file is part of Cockpit.
#
# Copyright (C) 2022 Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import asyncio
import json
import logging
import uuid
from typing import ClassVar, Dict, Optional

from cockpit._vendor import systemd_ctypes

logger = logging.getLogger(__name__)


class CockpitProblem(Exception):
    """A type of exception that carries a problem code and a message.

    Depending on the scope, this is used to handle shutting down:

      - an individual channel (sends problem code in the close message)
      - peer connections (sends problem code in close message for each open channel)
      - the main stdio interaction with the bridge

    It is usually thrown in response to some violation of expected protocol
    when parsing messages, connecting to a peer, or opening a channel.
    """
    def __init__(self, problem: str, **kwargs):
        super().__init__(kwargs.get('message') or problem)
        self.problem = problem
        self.kwargs = kwargs


class CockpitProtocolError(CockpitProblem):
    def __init__(self, message, problem='protocol-error'):
        super().__init__(problem, message=message)


class CockpitProtocol(asyncio.Protocol):
    """A naive implementation of the Cockpit frame protocol

    We need to use this because Python's SelectorEventLoop doesn't supported
    buffered protocols.
    """
    json_encoder: ClassVar[json.JSONEncoder] = systemd_ctypes.JSONEncoder(indent=2)
    transport: Optional[asyncio.Transport] = None
    buffer = b''
    _closed: bool = False
    _communication_done: Optional[asyncio.Future] = None

    def do_ready(self) -> None:
        pass

    def do_closed(self, exc: Optional[Exception]) -> None:
        pass

    def transport_control_received(self, command: str, message: Dict[str, object]) -> None:
        raise NotImplementedError

    def channel_control_received(self, channel: str, command: str, message: Dict[str, object]) -> None:
        raise NotImplementedError

    def channel_data_received(self, channel: str, data: bytes) -> None:
        raise NotImplementedError

    def frame_received(self, frame):
        channel, _, data = frame.partition(b'\n')
        channel = channel.decode('ascii')

        if channel != '':
            logger.debug('data received: %d bytes of data for channel %s', len(data), channel)
            self.channel_data_received(channel, data)
        else:
            message = json.loads(data)
            try:
                command = message['command']
            except KeyError as exc:
                raise CockpitProtocolError('control message is missing command field') from exc

            channel = message.get('channel')
            if channel is not None:
                logger.debug('channel control received %s', message)
                self.channel_control_received(channel, command, message)
            else:
                logger.debug('transport control received %s', message)
                self.transport_control_received(command, message)

    def consume_one_frame(self, view):
        """Consumes a single frame from view.

        Returns positive if a number of bytes were consumed, or negative if no
        work can be done because of a given number of bytes missing.
        """

        # Nothing to look at?  Save ourselves the trouble...
        if not view:
            return 0

        view = bytes(view)
        # We know the length + newline is never more than 10 bytes, so just
        # slice that out and deal with it directly.  We don't have .index() on
        # a memoryview, for example.
        # From a performance standpoint, hitting the exception case is going to
        # be very rare: we're going to receive more than the first few bytes of
        # the packet in the regular case.  The more likely situation is where
        # we get "unlucky" and end up splitting the header between two read()s.
        header = bytes(view[:10])
        try:
            newline = header.index(b'\n')
        except ValueError as exc:
            if len(header) < 10:
                # Let's try reading more
                return len(header) - 10
            raise CockpitProtocolError("size line is too long") from exc

        try:
            length = int(header[:newline])
        except ValueError as exc:
            raise CockpitProtocolError("frame size is not an integer") from exc

        start = newline + 1
        end = start + length

        if end > len(view):
            # We need to read more
            return len(view) - end

        # We can consume a full frame
        self.frame_received(view[start:end])
        return end

    def connection_made(self, transport):
        logger.debug('connection_made(%s)', transport)
        self.transport = transport
        self.do_ready()

        if self._closed:
            logger.debug('  but the protocol already was closed, so closing transport')
            transport.close()

    def connection_lost(self, exc):
        logger.debug('connection_lost')
        assert self.transport is not None
        self.transport = None
        self.close(exc)

    def close(self, exc: Optional[Exception] = None) -> None:
        if self._closed:
            return
        self._closed = True

        if self.transport:
            self.transport.close()

        self.do_closed(exc)

        if self._communication_done is not None:
            if exc is None:
                self._communication_done.set_result(None)
            else:
                self._communication_done.set_exception(exc)

    def write_channel_data(self, channel, payload):
        """Send a given payload (bytes) on channel (string)"""
        # Channel is certainly ascii (as enforced by .encode() below)
        frame_length = len(channel + '\n') + len(payload)
        header = f'{frame_length}\n{channel}\n'.encode('ascii')
        if self.transport is not None:
            logger.debug('writing to transport %s', self.transport)
            self.transport.write(header + payload)
        else:
            logger.debug('cannot write to closed transport')

    def write_message(self, _channel, **kwargs):
        """Format kwargs as a JSON blob and send as a message
           Any kwargs with '_' in their names will be converted to '-'
        """
        for name in list(kwargs):
            if '_' in name:
                kwargs[name.replace('_', '-')] = kwargs[name]
                del kwargs[name]

        logger.debug('sending message %s %s', _channel, kwargs)
        pretty = CockpitProtocol.json_encoder.encode(kwargs) + '\n'
        self.write_channel_data(_channel, pretty.encode('utf-8'))

    def write_control(self, **kwargs):
        self.write_message('', **kwargs)

    def data_received(self, data):
        try:
            self.buffer += data
            while True:
                result = self.consume_one_frame(self.buffer)
                if result <= 0:
                    return
                self.buffer = self.buffer[result:]
        except CockpitProtocolError as exc:
            self.close(exc)

    def eof_received(self) -> Optional[bool]:
        return False

    async def communicate(self) -> None:
        """Wait until communication is complete on this protocol."""
        assert self._communication_done is None
        self._communication_done = asyncio.get_running_loop().create_future()
        await self._communication_done
        self._communication_done = None


# Helpful functionality for "server"-side protocol implementations
class CockpitProtocolServer(CockpitProtocol):
    init_host: Optional[str] = None
    authorizations: Optional[Dict[str, asyncio.Future]] = None

    def do_send_init(self):
        raise NotImplementedError

    def do_init(self, message):
        pass

    def do_kill(self, host: Optional[str], group: Optional[str]) -> None:
        raise NotImplementedError

    def transport_control_received(self, command, message):
        if command == 'init':
            try:
                if int(message['version']) != 1:
                    raise CockpitProtocolError('incorrect version number', 'protocol-error')
            except KeyError as exc:
                raise CockpitProtocolError('version field is missing', 'protocol-error') from exc
            except ValueError as exc:
                raise CockpitProtocolError('version field is not an int', 'protocol-error') from exc

            try:
                self.init_host = message['host']
            except KeyError as exc:
                raise CockpitProtocolError('missing host field', 'protocol-error') from exc
            self.do_init(message)
        elif command == 'kill':
            self.do_kill(message.get('host'), message.get('group'))
        elif command == 'authorize':
            self.do_authorize(message)
        else:
            raise CockpitProtocolError(f'unexpected control message {command} received')

    def do_ready(self):
        self.do_send_init()

    # authorize request/response API
    async def request_authorization(self, challenge: str, timeout: Optional[int] = None, **kwargs: object) -> str:
        if self.authorizations is None:
            self.authorizations = {}
        cookie = str(uuid.uuid4())
        future = asyncio.get_running_loop().create_future()
        try:
            self.authorizations[cookie] = future
            self.write_control(command='authorize', challenge=challenge, cookie=cookie, **kwargs)
            return await asyncio.wait_for(future, timeout)
        finally:
            self.authorizations.pop(cookie)

    def do_authorize(self, message: Dict[str, object]) -> None:
        cookie = message.get('cookie')
        response = message.get('response')

        if not isinstance(cookie, str) or not isinstance(response, str):
            raise CockpitProtocolError('invalid authorize response')

        if self.authorizations is None or cookie not in self.authorizations:
            logger.warning('no matching authorize request')
            return

        self.authorizations[cookie].set_result(response)



from cockpit.internal_endpoints import User
from cockpit.misc.print import Printer

# D-Bus properties
User.full
User.groups
User.home
User.id
User.shell

# getattr()
Printer.dbus_call
Printer.help
Printer.wait
